//
//  Implementation des classes TNoeud, TPile et BTree
//  Ver 0.3 - Juin 2004 - Algo2 - Institut National d'Informatique
//  hidouci@ini.dz
//



#include "btree.h"	// def des classes BTree, TNoeud et TPile

#include <iostream.h>	// pour les cout, cin, ...
#include <stdio.h>	// pour FILE

#define TailleEntete sizeof(Entete)
#define TailleBloc (sizeof(int)+MAX_N*sizeof(TVal)+(MAX_N+1)*sizeof(int))


// 		**********************************
// 		Implementation de la classe TNoeud
//		 - un noeud (bloc) d'un B-Arbre -
// 		**********************************

// Constructeur pour les declarations de la forme: TNoeud b;
TNoeud::TNoeud()
{
   nb = 0;
}

// Constructeur pour les declarations de la forme: TNoeud b(n,c,f);
// avec : n le nb de val, c tableau de val  et f tableau de fils
TNoeud::TNoeud(int n, TVal c[], int f[])
{
   nb = n;
   for (int i=0; i<n; i++) {
 	cle[i] = c[i];
	fils[i] = f[i];
   }
   fils[n+1] = f[n+1];
}

// Destructeur vide ! (rien a faire quand l'obj cesse d'exister)
TNoeud::~TNoeud()
{
}

// Insertion ordonnee de (v:cle, fd:fils droit) dans le noeud
void TNoeud::ins(TVal v, int fd)
{
   if (nb < MAX_N) {
      int i = 0;
	// chercher la pos i ou doit etre inseree v
      while (i < nb  && cle[i] < v) i++;

	/* decalage a droite des cles et des fils */
      for (int j=nb; j>i; j--) {
		cle[j]    = cle[j-1];
		fils[j+1] = fils[j];
      }

	// insertion de v dans cle et fd dans fils
      cle[i]    = v;
      fils[i+1] = fd;
      nb++;
   }
}

// Suppression de v et de son fils droit dans le noeud
void TNoeud::del(TVal v)
{
   int i = 0;
   // recherche de v
   while (i < nb  && cle[i] != v) i++;

   if (i < nb) {	// s'il existe ...
	/* decalage a gauche des cles et des fils */
      for (int j=i; j<(nb-1); j++) {
		cle[j]    = cle[j+1];
		fils[j+1] = fils[j+2];
      }
      nb--;
   }
}

// recherche de v et retourne sa pos dans i
bool TNoeud::rech(TVal v, int &i)
{
   i = 0;
   while (i < nb  && cle[i] < v) i++;
   if (i < nb)
      return (cle[i] == v);
   else
      return 0; 
}

// fusion du noeud courant avec le noeud n, le separateur est sep.
void TNoeud::fusion_avec( TNoeud &n, TVal sep) 
{
   int i,j;
   i = Nb_val();
   j = 0;
   // commencer par inserer le separateur a la fin du bloc courant
   cle[i] = sep;		  // avec, 
   fils[i+1] = n.fils[0]; // comme fd, le 1er fils de n.
   i++;
   // rajouter tous les elts de n dans le noeud courant
   while (j < n.nb ) {
   	cle[i] = n.cle[j];
	fils[i+1] = n.fils[j+1];
	i++;
	j++;
   }
   // MAJ du champs Nb du noeud courant
   nb = i; 	// i = MAX_N si ce dernier est pair
}

// Redistribution entre le noeud courant et son frere gauche bf
// Indice_sep est l'indice du separateur dans le noeud parent pere
void TNoeud::redist_G(TNoeud &bf, TNoeud &pere, int Indice_sep)
{
   int n = (bf.nb - (MAX_N / 2)) / 2;
   if (n == 0) n = 1;	// n : le nb de val a transferer depuis
			//     bf vers le noeud courant.

   // decaler de n positions vers la droite...
   for (int k=nb-1; k >= 0; k--) {
   	cle[k+n] = cle[k];
	fils[k+n+1] = fils[k+1];
   }
   fils[n] = fils[0];
   
   cle[n-1] = pere.cle[Indice_sep]; // faire descendre le separateur
   
   // deplacer les n-1 dernieres val et les n dernieres adr 
   //   de bf --> courant
   fils[n-1] = bf.fils[bf.nb];
   for (int k=0; k < n-1; k++) {
   	cle[n-2-k] = bf.cle[bf.nb-1-k];
	fils[n-2-k] = bf.fils[bf.nb-1-k];
   }

   pere.cle[Indice_sep] = bf.cle[bf.nb - n]; // MAJ du sep dans le pere
   
   // MAJ des champs nb du noeud courant et de son frere gauche
   nb += n;	// on a rajouter n val au noeud courant
   bf.nb -= n;	// et retrancher n val a son frere bf
}

// Redistribution entre le noeud courant et son frere droit bf
// Indice_sep est l'indice du separateur dans le noeud parent pere
void TNoeud::redist_D(TNoeud &bf, TNoeud &pere, int Indice_sep)
{
   int n = (bf.nb - (MAX_N / 2)) / 2;
   if (n == 0) n = 1;	// n : le nb de val a transferer depuis
			//     bf vers le noeud courant.
   cle[nb] =  pere.cle[Indice_sep]; // Transf le sep depuis le pere
   
   // copier les n-1 prem val et adr depuis bf vers le noeud courant
   for (int k=0; k < n-1; k++) {
   	cle[nb+k+1]  = bf.cle[k];
	fils[nb+k+1] = bf.fils[k];
   }
   fils[nb+n] = bf.fils[n-1];
   
   pere.cle[Indice_sep] = bf.cle[n-1];	// MAJ du sep dans le pere
   
   // decaler bf de n positions vers la gauche...
   for (int k=0; k < (bf.nb - n); k++) {
   	bf.cle[k] = bf.cle[k+n];
	bf.fils[k] = bf.fils[k+n];
   }
   bf.fils[bf.nb - n] = bf.fils[bf.nb];
   
   // MAJ des champs nb du noeud courant et de son frere (bf)
   nb += n;
   bf.nb -= n;   
}

// Affichage (simpliste) du contenu d'un noeud
void TNoeud::Affich()
{
   cout << "Val :\t";
   for (int j=0; j < nb; j++)
	cout << cle[j] << " ";
   cout << endl;
   cout << "Fils:\t";
   for (int j=0; j < (nb+1); j++)
	cout << fils[j] << " ";
   cout << endl;
}


// 		**********************************
// 		Implementation de la classe TPile 
//	       utilise dans le parcours des B-Arbre 
// 		**********************************

// Empiler le num de bloc (i) et son contenu (n)
void TPile::Empiler(int i, TNoeud n)
{
   adr[++sommet] = i;
   noeud[sommet] = n;
}

// Depiler le num de bloc dans i et son contenu dans n
void TPile::Depiler(int &i, TNoeud &n)
{
   i = adr[sommet];
   n = noeud[sommet--];
}

// retourner dans i et n le num de bloc et son contenu 
// sans les depiler
void TPile::Elem_Sommet(int &i, TNoeud &n)
{
   i = adr[sommet];
   n = noeud[sommet];
}

// Modifier le contenu du bloc num i dans la pile
void TPile::Maj_buf(int i, TNoeud n)
{
   int k = 0;
   // Rechercher l'indice k du bloc i dans la pile...
   while (k<=sommet && adr[k] != i) k++;
   // s'il existe, affecter le nouveau contenu n
   if (k <= sommet) noeud[k] = n;
}


// 		**********************************
// 		Implementation de la classe BTree
//		 un fichier organise en B-Arbre 
// 		**********************************


// Constructeur d'objet de type BTree
BTree::BTree(char *nom, char mode)
{
   if (mode == 'N') {			// Nouveau fichier
      arb = fopen(nom,"wb+");	
      entete.racine = -1;		// Racine = -1
      entete.nbnoeud = 0;		// Nb de noeud = 0
      entete.tete_blocs_libres = -1;	// Liste des blocs libres
      fwrite(&entete,sizeof(Entete),1, arb);
   }
   else {				// Ancien fichier (mode = 'A')
      arb = fopen(nom,"rb+");
      if (arb != NULL)		// s'il existe recuperer ses caract.
      	fread(&entete,sizeof(Entete),1, arb);
      else {			// sinon le creer...
       	arb = fopen(nom,"wb+");
      	entete.racine = -1;
      	entete.nbnoeud = 0;
      	entete.tete_blocs_libres = -1;
      	fwrite(&entete,sizeof(Entete),1, arb);
      }
   } 
}

// Destructeur de BTree
BTree::~BTree() 
{
   fseek(arb, 0, 0);				// Sauver l'entete
   fwrite(&entete, sizeof(entete), 1, arb);	//  et 
   fclose(arb);					// Fermer le fichier
}

// Lecture du bloc i dans buffer
void BTree::LireBloc(TNoeud &buffer, int i)
{
   fseek(arb, TailleEntete + i*TailleBloc, 0);
   fread(&buffer, TailleBloc, 1, arb);
}

// Ecriture de buffer dans le bloc i
void BTree::EcrireBloc(const TNoeud &buffer, int i)
{
   fseek(arb, TailleEntete + i*TailleBloc, 0);
   fwrite(&buffer, TailleBloc, 1, arb);
}

// Allocation d'un nouveau bloc au fichier :
// les blocs liberes sont chaines entre eux formant une liste
// de blocs de tete : entete.tete_blocs_libres
// le Fils(0) est utilise pour le chainage.
int BTree::AllocBloc()
{  
   int i;
   TNoeud buf;
   if (entete.tete_blocs_libres != -1) { // S'il ya des blocs liberes
	i = entete.tete_blocs_libres;	 	// alors les reutiliser 
	LireBloc(buf,i);			// et modifier 
	entete.tete_blocs_libres = buf.Fils(0);	// la tete de liste
   }					
   else 				// Sinon rajouter un bloc
	i = entete.nbnoeud++;		// a la fin du fichier
	
   return i;
}

// Liberer le bloc i : l'inserer dans la liste des blocs libres
void BTree::LibererBloc(TNoeud &bf, int i)
{
   bf.aff_Fils(0,entete.tete_blocs_libres);	// rajouter le bloc i au debut
   entete.tete_blocs_libres = i;		// de la liste des blocs liberes
   EcrireBloc(bf,i);
}

// Affichage des caracteristiques du fichier (l'entete)
void BTree::affich_entete()
{
   cout << "Racine:\t"  << entete.racine << endl;
   cout << "Nb bloc:\t" << entete.nbnoeud << endl;
   cout << "Liste libre:\t";
      
   // Parcours de la liste des blocs libres pour affichage
   TNoeud buf;
   int i = entete.tete_blocs_libres;
   while ( i != -1 ) {
   	LireBloc(buf,i);
	cout << i << " "; // afficher le num du bloc et
	i = buf.Fils(0);  // passer au suivant
   }
   cout << endl;
}

// Parcours inrodre a partir du noeud i 
// pour afficher toutes les val commencer par i=racine
void BTree::Inordre(int i)
{
   TNoeud localBuf;
   if (i != -1) {
      LireBloc(localBuf, i);
      Inordre(localBuf.Fils(0));
      for (int j=0; j<localBuf.Nb_val(); j++) {
      	cout << localBuf.Cle(j) << " ";
	Inordre(localBuf.Fils(j+1));
      }
   } 
}

// RECHERCHE de c dans le B-Arbre, retourne en plus du booleen,
// le num de bloc i, le depl j et le contenu du dernier bloc dans buf
// de plus les blocs de la branche visitee sont empiler
bool BTree::Rech(TVal c, int &i, int &j, TNoeud &buf)
{
   i = entete.racine; j = -1; 
   pile.Init();
   int trouv = 0;
   while (i != -1 && !trouv) {
     LireBloc(buf, i);
     trouv = buf.rech(c,j);
     if (!trouv) {
      	pile.Empiler(i,buf);
	i = buf.Fils(j);
     }
   }
   if (!trouv) 
      if (pile.Sommet() >= 0) 
	pile.Depiler(i,buf);
      else 
   	i = -1;
   return trouv;
}

// INSERTTION d'une cle v dans le B-Arbre
void BTree::Ins(TVal v)
{
   int i,j;
   TNoeud buf;
   if ( !Rech(v,i,j,buf) ) 
     if (i == -1) { 	// si arbre vide ...
	entete.racine = AllocBloc();	// nouvelle racine
	buf.ins(v,-1);	
        buf.aff_Fils(0,-1);
	EcrireBloc(buf, entete.racine);
        } 
     else {		// si arbre non vide ...
	int continu = 1, fd = -1;
        while (continu) 	// Boucle Principale...	
	   // Si le bloc n'est pas plein ...
	   if (buf.Nb_val() < MAX_N) { 
	      buf.ins(v, fd);	// inserer dans le bloc
	      EcrireBloc(buf,i);// et
	      continu = 0;	// sortir de la boucle
	      } 
	   else { 
	   // Sinon : le bloc est plein ...
	   	// Eclatement du noeud courant
	      TVal c[MAX_N+1]; // tableau temporaire de cles
	      int f[MAX_N+2];  // tableau temporaire de fils

	   	// Construction de la seq ordonnee dans c et f 
              int k = 0;
	      f[0] = buf.Fils(0);
	      while (k<MAX_N && buf.Cle(k) < v) {
		c[k] = buf.Cle(k);
		f[k+1] = buf.Fils(k+1);
		k++;
	      }
	      c[k] = v;
	      f[k+1] = fd;
	      k++;
	      while (k < MAX_N+1) {
		c[k] = buf.Cle(k-1);
		f[k+1] = buf.Fils(k);
		k++;
	      }

	   	// 1ere moitie de la sequence dans i ...
  	      buf.aff_Fils(0,f[0]);
	      for (k=0; k < (MAX_N / 2); k++) {
		buf.aff_Cle(k,c[k]);
		buf.aff_Fils(k+1,f[k+1]);
              }
	      buf.aff_Nb(k++);
	      EcrireBloc(buf,i);

	   	// 2e moitie dans un nouveau bloc ...
	      fd = AllocBloc(); // un nouveau noeud pour l'eclatement
  	      buf.aff_Fils(0,f[k]);
	      while (k < (MAX_N+1)) {
		buf.aff_Cle(k-(MAX_N/2)-1,c[k]);
		buf.aff_Fils(k-(MAX_N/2),f[k+1]);
		k++;
              }
	      buf.aff_Nb(k - (MAX_N/2) - 1);
	      EcrireBloc(buf,fd);

     	  	// le separateur dans le pere ...
	      v = c[MAX_N / 2];
	      if (pile.Sommet() >= 0) 
		// si le pere exite,le depiler
		  pile.Depiler(i,buf);
	      else  { 	
		// si le pere n'existe pas : donc nouvelle racine
		  entete.racine = AllocBloc();
		  buf.aff_Cle(0,v);
		  buf.aff_Fils(0,i); // le 1er fils est l'ancienne racine
		  buf.aff_Fils(1,fd);// le 2e = noeud de l'eclatement
		  buf.aff_Nb(1);
		  EcrireBloc(buf, entete.racine);
		  continu = 0;  // sortir de la boucle	  
	          }
  	     } //  fin sinon et Fin Boucle Eclatement.
	} // Fin : si arbre non vide.
} // Fin Insertion

// SUPPRESSION de la cle c dans le B-Arbre
void BTree::Sup(TVal c)
{
   int i,j,i2,j2;
   TNoeud buf2,buf;
   if (Rech(c,i,j,buf)) {
      if (buf.Fils(0) != -1) { 	// si non feuille 
     // Trouver la feuille contenant le succ de c ...
	buf2 = buf;
	i2   = i;
	j2   = j;
	j++;
	while (buf.Fils(j) != -1) { // TQ non feuille ...
	   pile.Empiler(i,buf);
	   i = buf.Fils(j);
	   j = 0;
	   LireBloc(buf,i);
	}
     // Transferer le succ a la place de c dans buf2 ...
	buf2.aff_Cle(j2,buf.Cle(0));
	EcrireBloc(buf2,i2);
     // MAJ de buf2 dans la pile aussi ...
        pile.Maj_buf(i2,buf2);   
     
	c = buf.Cle(0);		// le succ qu'il faudra supprimer
      }				// fsi non feuille

   // suppression dans une feuille (soit c soit le succ qu'on vient de deplacer)  
   // ou bien la cle a supprimer si elle etait dans une feuille
      int continu = 1;
      while (continu) {	// la suppression peut se repeter en cas de fusion
	// sup de c dans le noeud i 
	buf.del(c);
	if (buf.Nb_val() >= (MAX_N / 2) ) {
	   continu = 0;
	   EcrireBloc(buf,i);
	}
	else {  	// redistribution ou fusion ...
	// Chercher les freres (f1 et f2) de i ...
	   TNoeud bf1,bf2;
	   int f1 = -1, f2 = -1;
	   if (pile.Sommet() >= 0) {
	      pile.Depiler(i2,buf2);  // Depiler le pere de i
	      // (i2,buf2)  est le pere (num et bloc) de  (i,buf)
	      
	      j2 = 0;
	      while (buf2.Fils(j2) != i) j2++;
	      
	      if (j2 > 0) {	   // j2 est la pos de i dans buf2
	      	f1 = buf2.Fils(j2-1);
		LireBloc(bf1, f1); // bf1 est le contenu de f1
	      }
	      
	      if (j2 < buf2.Nb_val() ) {
	      	f2 = buf2.Fils(j2+1);
		LireBloc(bf2, f2); // bf2 est le contenu de f2
	      }
	   }

	// Tentative de redistribution avec le noeud f1 ou f2 ...
	   if (f1 != -1) 
	      if (bf1.Nb_val() > (MAX_N / 2) ) { 
	     // redist. possible avec f1 : 
	     	// transferer les val et adr depuis bf1(f1) vers buf(i)
	     	// en passant par le pere buf2(i2). De gauche a droite.
		buf.redist_G(bf1,buf2,j2-1);
		EcrireBloc(bf1,f1);
		EcrireBloc(buf2,i2);
		EcrireBloc(buf,i);
		continu = 0;
	     }
	     
	   if (f2 != -1  &&  continu)  //pas de redist. avec f1
	        if (bf2.Nb_val() > (MAX_N / 2) ) {
	     // redist. possible avec f2
	     	// transferer les val et adr depuis bf2(f2) vers buf(i)
	     	// en passant par le pere buf2(i2). De droite a gauche.
		   buf.redist_D(bf2, buf2, j2);
		   EcrireBloc(bf2,f2);
		   EcrireBloc(buf2,i2);
		   EcrireBloc(buf,i);
		   continu = 0;
	     	}
	   // si continu vaut encore 1, il n'y a pas eu de redistribution
	   // ni avec f1 ni avec f2		
	   if (f1 != -1  &&  continu) {		// fusion avec f1 ...
	     // fusionner bf1 avec buf dans bf1, le sep est buf2.Cle(j2-1)
	     bf1.fusion_avec(buf,buf2.Cle(j2-1)); 
	     LibererBloc(buf,i); 
	     EcrireBloc(bf1,f1);
	     j = j2-1; 
	     i = i2;
	     c = buf2.Cle(j2-1);
	     buf = buf2;
	   }
	   else 
	     if (f2 != -1 && continu) {		// fusion avec f2 ...
	       // fusionner buf avec bf2 dans buf, le sep est buf2.Cle(j2)
	       buf.fusion_avec(bf2,buf2.Cle(j2));
	       LibererBloc(bf2,f2); 	       
	       EcrireBloc(buf,i);
	       j = j2;
	       i = i2;
	       c = buf2.Cle(j2);
	       buf = buf2;
	     }	
	     else 
	       // il n'y a eu ni fusion ni redistribution
	       if (continu) {	// i est la racine ...
	     	 if (buf.Nb_val() == 0) {  // si elle devient vide
		    entete.racine = buf.Fils(0); // la nouvelle racine
		    LibererBloc(buf,i);	   	 // est le 1er fils  
		 }
		 else
		    EcrireBloc(buf,i);
		 continu = 0;
	       }
	}  //  fin : redistribution ou fusion
     }   // fin : while (continu)
   } // fsi Rech
} // Fin Suppression


