//
//  Definition des classes BTree, TNoeud et TPile
//  Ver 0.3 - Juin 2004 - Algo2 - Institut National d'Informatique
//  hidouci@ini.dz
//

#ifndef BTREE_H
#define BTREE_H

#include <stdio.h>	// pour la structure FILE

#define MAX_N 4		// Taille d'un noeud = nb max de cles par bloc
#define MAX_P 10	// Taille de la pile ( > Profondeur max)

typedef int TVal;		// La cle ou bien enregistrement

typedef struct bloc0 {		// L'entete du fichier B-Arbre :
	int racine;		// num de bloc racine
	int tete_blocs_libres;	// Liste des blocs liberes (pour recuperation)
	int nbnoeud;		// Nombre de blocs alloues au fichier
	} Entete;
	

// 		******************************
//		Definition de la classe TNoeud
//		 un noeud (bloc) d'un B-Arbre  
// 		******************************

class TNoeud {		// un Bloc du B-Arbre :
private:
   int nb;		// nb de valeurs (cles) presentes dans le bloc
   TVal cle[MAX_N];	// Tableau des valeurs (cles)
   int fils[MAX_N+1];	// Tableau d'adr vers les noeuds fils

public:
   TNoeud();				// constructeur d'un noeud vide
   TNoeud(int n, TVal c[], int f[]);	// constructeur d'un noeud non vide
   ~TNoeud();				// destructeur
// insere une valeur avec son fils droit   
   void ins(TVal v, int fd);
// supprime une valeur et son fils droit   
   void del(TVal v);
// recherche la pos i d'une valeur v   
   bool rech(TVal v, int &i);	
// fusion du noeud courant avec le noeud n, le separateur est sep.
   void fusion_avec( TNoeud &n, TVal sep); 
// Redistribution entre le noeud courant et son frere gauche bf
   void redist_G(TNoeud &bf, TNoeud &pere, int Indice_sep); 
// Redistribution entre le noeud courant et son frere droit bf
   void redist_D(TNoeud &bf, TNoeud &pere, int Indice_sep); 
// Affichage du noeud   
   void Affich();			   
// methodes implementees en inline...
   int Nb_val() { return nb; }		// retourne le nb de val dans le noeud
   TVal Cle(int i) { return cle[i]; }	// retourne la i eme val 
   int Fils(int i) { return fils[i]; }	// retourne le i eme fils (adr)
   void aff_Cle(int i, TVal v) { cle[i] = v; }	// modif la i eme val
   void aff_Fils(int i, int j) { fils[i] = j; } // modif le i eme fils
   void aff_Nb(int i) { nb = i; }		// modif le nb de val
};


// 		******************************
//		Definition de la classe TPile
//	      utilise dans le parcours du B-Arbre 
// 		******************************

class TPile {		// une pile utilisee 
private:
   int sommet;		// pour garder la trace
   int adr[MAX_P];	// des adresses et
   TNoeud noeud[MAX_P];	// des noeuds visites

public:
   TPile() { sommet = -1; }		// inline
   int Sommet() { return sommet; }	// inline
   void Init() { sommet = -1; }		// inline
   
   void Empiler(int i, TNoeud n);
   void Depiler(int &i, TNoeud &n);
   void Elem_Sommet(int &i, TNoeud &n);  // le dernier element sans le depiler
   void Maj_buf(int i, TNoeud n);	 // modifie le noeud d'adresse i dans la pile
};


// 		******************************
//		Definition de la classe BTree
//		   un Fichier en B-Arbre 
// 		******************************

class BTree {	// Un Fichier organise en B-Arbre
private:
   FILE *arb;	// Le fichier contenant les noeuds
// les caracteristiques du B-Arbre :   
   Entete entete;	// (racine, nb_noeud, ...)
// Pile de parcours (garde la trace des blocs visites)   
   TPile pile;		

public:
// Constructeur et destructeur de la classe
   BTree(char *nom, char mode = 'A'); 	// mode ='A' (defaut) ou 'N'
   ~BTree();				 

// Recherche de c dans le B-Arbre et retourne dans i,j le num de bloc
// et le deplacement et dans buf le dernier noeud visite. 
   bool Rech(TVal c, int &i, int &j, TNoeud &buf); 
   	
   void Ins(TVal c);		// insertion d'une valeur (c)
   void Sup(TVal c);		// suppression d'une valeur (c)
// Affichage de toutes les valeurs (triees) a partir du noeud i   
   void Inordre(int i);		
// lire le bloc i dans la var buffer   
   void LireBloc(TNoeud &buffer, int i);
// ecrire la var buffer dans le bloc i   
   void EcrireBloc(const TNoeud &buffer, int i);
// Alloue et retourne le num d'un nouveau bloc   
   int AllocBloc();		
// libere le bloc i   
   void LibererBloc(TNoeud &bf, int i);
// affichage des caract. du fichier
   void affich_entete();   
// retourne le num du bloc racine (methode inline) 
   int Racine() { return entete.racine; }
};


#endif

// L'implementation des methodes (non inline) est dans : btree.cpp

